#ProcessorSurveys.py
'''Functions to process the survey component.

Group surveys and run source gen on each groups:
Grouping is only run if it is defined in the config (otherwise it use all available surveys).
Example:
    "Groups" : {
                "PROD" : ["ODK-server", "kobo-server", "CSV-CAM"],
                "WRT" : ["ODK", "CSV"]
            } 

It is based on the 'Origin' column of the surveys.
All unique 'Origin' will create its own group alongside the ones defined in the config.
Groups can be ignored by defining 'GroupsToIgnore' in the config (this has no effect on 'Origin' types).
Example:
    "GroupsToIgnore" : ["ODK-server", "kobo-server", "CSV-CAM", "ODK", "CSV", "PV"]
'''

import csv
import datetime
from glob import glob
import json
import logging
from pathlib import Path
import os

from numpy import all as np_all
from numpy import any as np_any
from shutil import copyfile
from pandas import read_csv, concat

from Processor import Processor
from source_gen.clustering import run_case

from ProcessorSurveysODK import get_ODK_form_as_csv
from ProcessorSurveysODKv2 import get_ODKv2_form_as_csv
from ProcessorSurveysnewODK import get_newODK_form_as_csv
from ProcessorSurveysnewODK2 import get_newODK2_form_as_csv
from ProcessorSurveyskobotoolbox import get_kobotoolbox_form_as_csv
from ProcessorSurveysWRSIS import get_WRSIS_form_as_csv
from ProcessorSurveysWRT import get_WRT_form_as_csv
from ProcessorUtils import (
        subprocess_and_log,
        endJob,
        add_filters_to_sublogger,
)


class ProcessorSurveys(Processor):

    def process_pre_job(self, args):
        return self.process_pre_job_survey(args)


    def process_in_job(self, jobPath, status, configjson, component) -> object:
        return self.process_in_job_survey(jobPath, status, configjson, component)


    def process_post_job(self, jobPath, configjson):
        return self.process_EWS_plotting_survey(jobPath, configjson)


    def __init__(self) -> None:
        super().__init__()
        logger = logging.getLogger('Processor.Surveys')
        add_filters_to_sublogger(logger)

        self.GET_FORM_AS_CSV_DICT = {
            'ODK':         get_ODK_form_as_csv,
            'kobotoolbox': get_kobotoolbox_form_as_csv,
            'WRSIS':       get_WRSIS_form_as_csv,
            'WRT':         get_WRT_form_as_csv,
            'ODKv2':       get_ODKv2_form_as_csv,
            'newODK':      get_newODK_form_as_csv,
            'newODK2' : get_newODK2_form_as_csv,
        }

    def process_pre_job_survey(self, input_args):
        '''Returns a boolean as to whether the job is ready for full processing.'''
        self.logger.info('started process_pre_job_survey(), nothing to do')

        return True

    def process_in_job_survey(self, jobPath,status,config,component):
        self.logger.info('started process_in_job_survey()')

        self.logger.debug('Performing download(s) from ODK server')

        credentials_filename = config['Survey']['ServerCredentialsFile']
        with open(credentials_filename) as credentials_file:

            cred: dict = json.load(credentials_file)

            assert 'forms' in cred.keys()

        csv_filenames = {}
        for form in cred['forms']:

            self.logger.debug(f"Starting to download {form['form_id']}")

            assert form['type'] in self.GET_FORM_AS_CSV_DICT

            func_get_form_as_csv = self.GET_FORM_AS_CSV_DICT[form['type']]

            csv_filename = func_get_form_as_csv(form, jobPath, config, status)

            csv_filenames[form['form_id']] = csv_filename

        # load each file of surveys as a dataframe
        forms = {}
        for form_name,form_fn in csv_filenames.items():

            # some define column types, hardwired for now
            col_types = {'comment':'str','KEY':'str'}

            form_df = read_csv(form_fn,dtype=col_types)

            forms[form_name] = form_df

        # create some standard dataframe modification functions
        def add_column(df,coln,value):
            df[coln]=value
            return

        def remove_column(df,coln,value):
            del df[coln]
            return

        def replace_column(df,coln,value):
            df[coln]=value
            return

        def filter_by_column(df,coln,value):
            # CAUTION: This requires surveyor to provide the correct country
            df.drop(df.loc[df[coln]!=value].index,inplace=True)
            #TODO : for Kenya data, provide a coordinate-based filter
            return

        def filter_by_list(df,coln,values):
            # CAUTION: This requires surveyor to provide the correct list of countries
            df.drop(df.loc[~df[coln].isin(values)].index,inplace=True)
            return

        func_types = {
            'add': add_column,
            'remove' : remove_column,
            'replace' : replace_column,
            'filter' : filter_by_column,
            'filter_by_list' : filter_by_list
        }

        # simple format alignment using edits on config
        # (should this need to be much more sophisticated, reconsider the workflow)
        if 'FormEdits' in config['Survey']:

            form_edits = config['Survey']['FormEdits']

            # loop over each form
            for form_name, edits in form_edits.items():

                form_df = forms[form_name]

                # loop over each type of edit
                for func_type, columns in edits.items():

                    # check the function is available
                    assert func_type in func_types

                    # loop over each column to modify
                    for coln,val in columns.items():

                        # apply the edit
                        func_types[func_type](form_df,coln,val)

        # Merge additional SurveyData files and rearrange columns to be consistent
        # Assumes that the same columns are present in all forms
        # and that the first form is the standard

        first=True
        for dfi in forms.values():

            if first:
                standard_columns = dfi.columns.tolist()
                dfm = dfi

                self.logger.debug(f"First processed form contains {dfm.shape[0]} records")

                first=False
                continue

            # re-order columns to match first case (presumed standard format)
            # and fill missing columns with empty strings
            dfi = dfi.reindex(standard_columns,fill_value='',axis='columns')

            self.logger.debug(f"Next processed form contains {dfi.shape[0]} records")

            dfm = concat([dfm,dfi],axis='rows')

        # save the result
        Export_csv_path = f"{jobPath}/ExportCSV/"
        Path(Export_csv_path).mkdir(parents = True, exist_ok = True)
        forms_fn = f"{Export_csv_path}/Merged_SurveyData.csv"
        dfm.to_csv(forms_fn,index=False,quoting=csv.QUOTE_MINIMAL)

        self.logger.debug(f"Preparing to apply removals and additions to survey data")

        processed_surveys_filepath = f"{Export_csv_path}/Processed_SurveyData.csv"

        survey_errors_to_remove_filepath = f"{config['ResourcesPath']}/coordinator/assets/SURVEYDATA_MANUAL/SurveyDataErrorsToRemove.csv"
        survey_additions_filepath = f"{config['ResourcesPath']}/coordinator/assets/SURVEYDATA_MANUAL/LIVE_SURVEYDATA_TOUSE.csv"

        # perform here in python, using the 'KEY' column
        # check the key column is unique, if not raise a warning and remove duplicates

        if dfm['KEY'].unique().size != dfm['KEY'].size:
            status.reset('WARNING')
            self.logger.warning(f"KEY column is not unique, removing duplicates")
            # count the number of duplicates
            n_duplicates = dfm.shape[0] - dfm['KEY'].unique().size
            # drop the duplicates
            dfm = dfm.drop_duplicates(keep='first')
            self.logger.warning(f"Removed {n_duplicates} duplicates")

        df_rm = read_csv(survey_errors_to_remove_filepath,dtype='str')
        keys_to_rm = df_rm['KEY']

        # check that all of the keys to remove exist in the original data
        rm_keys_found = df_rm['KEY'].isin(dfm['KEY'])
        n_rm_keys_found = rm_keys_found.sum()
        n_rm_keys = rm_keys_found.size
        if not np_all(rm_keys_found):
            # this might happen if the run date is in the past
            self.logger.warning(f"Only found {n_rm_keys_found} of {n_rm_keys} survey errors to remove")

            rm_keys_not_found = df_rm[~rm_keys_found]
            self.logger.debug(f"Erroneous entries not found are:\n{rm_keys_not_found}")

            self.logger.debug(f"Type of keys that can be found include:\n{dfm['KEY'].dtype}")

            dfm_short_keys = [val for val in dfm['KEY'].values if len(str(val)) <10]
            self.logger.debug(f"Keys that can be found include:\n{dfm_short_keys}")

        # identify which surveys to remove
        idx_to_rm = dfm['KEY'].apply(lambda cell: cell in keys_to_rm.values)

        #drop them in-place
        dfm = dfm[~idx_to_rm]
        self.logger.info(f"Removed {n_rm_keys_found} erroneous surveys")

        # add the extra entries
        df_add = read_csv(survey_additions_filepath,dtype='str')
        n_add_keys = df_add.shape[0]
        df_join = concat([dfm,df_add])
        assert dfm.shape[0]+df_add.shape[0] == df_join.shape[0], 'Unexpected result of including additional surveys'

        self.logger.info(f"Added {n_add_keys} additional surveys")

        # save as processed
        df_join.to_csv(processed_surveys_filepath,index=False,quoting=csv.QUOTE_MINIMAL)

        self.logger.debug('Preparing clustering calculation')

        date = datetime.datetime.now()

        # prepare environment for clustering calc
        call_R = False

        upload_directory = f"{jobPath}/upload"
        Path(upload_directory).mkdir(parents=True, exist_ok=True)

        if call_R:

            cluster_calc_path = "/storage/app/EWS_prod/code/wheat_source_generation/"

            # clear old output
            old_clustering_output_glob = f"{cluster_calc_path}/output/sources_*"
            old_clustering_outputs = glob(old_clustering_output_glob)

            self.logger.info('About to unlink old output from clustering calculation')
            for path in old_clustering_outputs:
                self.logger.info(f"unlinking {path}")
                Path(path).unlink()


            RPath = '/usr/local/R/bin/Rscript'

            clustering_script = f"{cluster_calc_path}/code/R/clustering.R"

            clustering_env = {
                    **os.environ,
                    'R_LIBS':'/home/ewsmanager/R-packages-EWS-clustering/x86_64-pc-linux-gnu-library/3.5',
                    'PROJ_LIB' : '/usr/share/proj/', # conda env breaks the automatic assignment of PROJ_LIB
                    }

            clustering_config = config['Survey']['SourcesConfigFilename']
            assert os.path.isfile(clustering_config)

            clustering_calc = [RPath,
                    '--no-init-file',
                    clustering_script,
                    processed_surveys_filepath,
                    config['StartString'],
                    '-2',
                    '7',
                    config['Survey']['SourcesConfigFilename']]

            self.logger.debug('Performing clustering calculation')

            description_short = 'wheat-source-generation'
            description_long = 'source calculation on processed surveys'

            try:
                subprocess_and_log(clustering_calc, description_short, description_long, env=clustering_env)
            except:
                status.reset('ERROR')
                endJob(status,premature=True)

            self.logger.debug('Checking output of clustering calculation')

            try:
                self.logger.debug('Trying to copy the dataset processed for clustering')

                clustering_proc_path_glob = f"{cluster_calc_path}/output/survey_data_processed_{config['Survey']['SourcesRegionName']}_{date.strftime('%Y-%m-%d')}_*.csv"
                clustering_proc_path_list = glob(clustering_proc_path_glob)
                if len(clustering_proc_path_list) == 0:
                    self.logger.debug(f"No processed files produced from clustering in {clustering_proc_path_glob}")
                    raise Exception

                elif len(clustering_proc_path_list) > 1:
                    self.logger.debug(f"Multiple processed files produced from clustering in {clustering_proc_path_glob}")
                    raise Exception

                else:
                    self.logger.debug('Found 1 processed file, placing copy of result in job directory')

                    proc_filename = f"survey_data_processed_{config['StartString']}.csv"
                    proc_path = f"{output_directory}/{proc_filename}"

                    self.logger.debug(f"as {proc_path}")

                    copyfile(clustering_proc_path_list[0], proc_path)

            except:
                self.logger.debug('Failed to get a copy of the dataset processed for clustering')

            clustering_output_path_glob = f"{cluster_calc_path}/output/sources_{config['Survey']['SourcesRegionName']}_{date.strftime('%Y-%m-%d')}_*.csv"
            clustering_output_path_list = glob(clustering_output_path_glob)
            if len(clustering_output_path_list) == 0:
                self.logger.error(f"No output produced from clustering in {clustering_output_path_glob}")
                status.reset('ERROR')
                endJob(status,premature=True)
            if len(clustering_output_path_list) > 1:
                self.logger.error(f"Multiple outputs produced from clustering in {clustering_output_path_glob}")
                status.reset('ERROR')
                endJob(status,premature=True)

            sources_path = clustering_output_path_list[0]

        elif 'Groups' in config['Survey']:
            # if 'Groups' is defined in the config, create grouped survey files and run python version

            self.logger.debug('Preparing grouped survey files')
            group_directory = f"{jobPath}/Groups"
            Path(group_directory).mkdir(parents=True, exist_ok=True)

            origins_list = df_join["Origin"].unique()
            groups = {i:[i] for i in origins_list}

            assert not np_any([k in origins_list for k in config['Survey']['Groups'].keys()])

            groups.update(config['Survey']['Groups'])

            # remove groups that are listed in GroupsToIgnore
            if 'GroupsToIgnore' in config['Survey']:
                for group_name in config['Survey']['GroupsToIgnore']:
                    if group_name in groups:
                        self.logger.info(f"Removing group {group_name} from list of groups")
                        del groups[group_name]

            for group_name,group_content in groups.items():

                self.logger.info(f"Creating survey group {group_name} which includes {group_content}")

                df_group = df_join.loc[df_join["Origin"].isin(group_content)]

                group_surveys_filename = f"surveys_{group_name}.csv"
                group_surveys_filepath = f"{group_directory}/{group_surveys_filename}"

                df_group.to_csv(group_surveys_filepath, index=False, quoting=csv.QUOTE_MINIMAL)

                output_directory = f"{jobPath}/source_gen/{group_name}"
                Path(output_directory).mkdir(parents=True, exist_ok=True)

                sources_path = run_case(
                        config_path = config['Survey']['pySourcesConfigFilename'],
                        survey_path = group_surveys_filepath,
                        reference_date = config['StartString'],
                        # Day 0 (current day) is always included
                        # Days -2,-1 and 0 are needed to initialise spores in NAME
                        # days 1-7 are forecast days
                        # days 8 and 9 are buffers in case NAME needs to run with the
                        # previous day's source job
                        day_offsets = [-2,9],
                        output_dir = output_directory)

                self.logger.debug('Placing copy of result in job directory with conventional name')

                output_filename = f"sources_{group_name}_{config['StartString']}.csv"
                output_path = f"{jobPath}/upload/{output_filename}"

                self.logger.debug(f"as {output_path}")

                copyfile(sources_path, output_path)
        else:
            # run python version without grouping surveys

            output_directory = f"{jobPath}/source_gen"
            Path(output_directory).mkdir(parents=True, exist_ok=True)

            sources_path = run_case(
                    config_path = config['Survey']['pySourcesConfigFilename'],
                    survey_path = processed_surveys_filepath,
                    reference_date = config['StartString'],
                    # Day 0 (current day) is always included
                    # Days -2,-1 and 0 are needed to initialise spores in NAME
                    # days 1-7 are forecast days
                    # days 8 and 9 are buffers in case NAME needs to run with the
                    # previous day's source job
                    day_offsets = [-2,9],
                    output_dir = output_directory)

            self.logger.debug('Placing copy of result in job directory with conventional name')

            output_filename = f"sources_{config['StartString']}.csv"
            output_path = f"{jobPath}/upload/{output_filename}"

            self.logger.debug(f"as {output_path}")

            copyfile(sources_path, output_path)


        upload_filenames = f"sources_*{config['StartString']}.csv"
        upload_path = f"{jobPath}/upload/{upload_filenames}"

        # glob list of output files
        upload_path_list = glob(upload_path)

        proc_out = {}
        # Output files available for upload
        proc_out['output'] = upload_path_list
        # Processing files available for clearing
        proc_out['clearup'] = None

        return proc_out

    #TODO
    def process_EWS_plotting_survey(self, jobPath,config):
        '''Returns a list of output files for transfer.'''

        self.logger.info('started process_EWS_plotting_survey(), nothing to do')

        pass
        return []


if __name__ == '__main__':
    processor = ProcessorSurveys()
    processor.run_processor("Survey")
